// Ovladani RGB WS2812 LED diody pro signalizacni ucely
/**The MIT License (MIT)
Copyright (c) 2017,20,23 by Pavel Brychta, http://www.xpablo.cz
Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:
The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.
THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
*/

#pragma once
#include <Arduino.h>
#include <Ticker.h>
#if defined(ARDUINO_ARCH_ESP32)
    #include <driver/rmt.h>
    #define BITS_PER_LED_CMD 24
    #define LED_BUFFER_ITEMS ((1 * BITS_PER_LED_CMD))
#endif

struct LEDRGB {
    union {
        struct {
            union {
                uint8_t r;
                uint8_t red;
            };
            union {
                uint8_t g;
                uint8_t green;
            };
            union {
                uint8_t b;
                uint8_t blue;
            };
        };
        uint8_t raw[3]{};
    };

    // Array access operator to index into the crgb object
    inline uint8_t & operator[](uint8_t x) __attribute__((always_inline))
    {
        return raw[x];
    }

    // Array access operator to index into the crgb object
    inline const uint8_t & operator[](uint8_t x) const __attribute__((always_inline))
    {
        return raw[x];
    }

    // default values are UNINITIALIZED
    inline LEDRGB() __attribute__((always_inline))
    {
    }

    // allow construction from R, G, B
    inline LEDRGB(uint8_t ir, uint8_t ig, uint8_t ib) __attribute__((always_inline))
        : r(ir), g(ig), b(ib)
    {
    }

    // allow construction from 32-bit (really 24-bit) bit 0xRRGGBB color code
    inline LEDRGB(uint32_t colorcode) __attribute__((always_inline))
        : r((colorcode >> 16) & 0xFF), g((colorcode >> 8) & 0xFF), b((colorcode >> 0) & 0xFF)
    {
    }

    // allow copy construction
    inline LEDRGB(const LEDRGB & rhs) __attribute__((always_inline))
    {
        r = rhs.r;
        g = rhs.g;
        b = rhs.b;
    }

    // allow assignment from one RGB struct to another
    inline LEDRGB & operator=(const LEDRGB & rhs) __attribute__((always_inline))
    {
        r = rhs.r;
        g = rhs.g;
        b = rhs.b;
        return *this;
    }

    // allow assignment from 32-bit (really 24-bit) 0xRRGGBB color code
    inline LEDRGB & operator=(const uint32_t colorcode) __attribute__((always_inline))
    {
        r = (colorcode >> 16) & 0xFF;
        g = (colorcode >> 8) & 0xFF;
        b = (colorcode >> 0) & 0xFF;
        return *this;
    }

    // allow assignment from R, G, and B
    inline LEDRGB & setRGB(uint8_t nr, uint8_t ng, uint8_t nb) __attribute__((always_inline))
    {
        r = nr;
        g = ng;
        b = nb;
        return *this;
    }
};

enum ColorOrder {
    RGB = 0012,
    GRB = 0102
};

typedef enum {
    LED_STOP = 0,
    LED_BLINK,
    LED_BREATH,
    LED_PULSE_START,
    LED_PULSE,
    LED_PATTERN,
} ledstate_t;

class wsLED {
  protected:
    int mPin; // pin, na kterem je LED pripojena
    LEDRGB mColor1;
    LEDRGB mColor2;
    int mOrder; // ColorOrder
    uint8_t mScale = 0xff;
    uint8_t mColor[3]{};
    Ticker mTimer;
    volatile ledstate_t mLedState = LED_STOP; // stav LEDky (blika, dycha, ...)
    volatile ledstate_t mSavedState = LED_STOP; // stav pro navrat
    volatile unsigned mSubState = 0; // pomocny stav/index v poli dychani
    uint32_t mPulseDuration = 0;
    uint32_t mBlinkSpeed = 0;
    LEDRGB mPulseColor;
#if defined(ARDUINO_ARCH_ESP32)
    int _rmtChannel = 0;
    rmt_item32_t led_data_buffer[LED_BUFFER_ITEMS] = {0};
#endif
    uint32_t mEndTime = 0;
    uint8_t mPrevColor[3] = {1, 2, 3}; // some not used colors

    void show();
    void setColor(const LEDRGB & color);
    [[nodiscard]] bool canShow() const;

    void rtLed(); // vykonna metoda

    static void ledTrampoline(wsLED * ptr) {

        ptr->rtLed();
    }

  public:
    explicit wsLED(int pin, int order = RGB);
    wsLED();
    void begin();
    void begin(int pin);
    void setColors(const LEDRGB & color1, const LEDRGB & color2 = LEDRGB(0, 0, 0));
    void blink(uint32_t speed = 300);
    void breath();
    void setOrder(int neworder);
    ledstate_t getState();
    void pulse(const LEDRGB & color, uint32_t duration);
    void stop(const LEDRGB & color = LEDRGB(0, 0, 0));
};
